
typedef union
#ifdef __cplusplus
	YYSTYPE
#endif
 {
	char	*str;
	int	num;
} YYSTYPE;
extern YYSTYPE yylval;
# define IL_NUMBER 257
# define IL_TOKEN 258
# define IL_HEXDIGIT 259
# define IL_COLON 260
# define IL_DOT 261
# define IL_EOF 262
# define IL_COMMENT 263
# define IL_INTERFACE 264
# define IL_IFNAME 265
# define IL_MTU 266
# define IL_EADDR 267
# define IL_IPV4 268
# define IL_V4PROTO 269
# define IL_V4SRC 270
# define IL_V4DST 271
# define IL_V4OFF 272
# define IL_V4V 273
# define IL_V4HL 274
# define IL_V4TTL 275
# define IL_V4TOS 276
# define IL_V4SUM 277
# define IL_V4LEN 278
# define IL_V4OPT 279
# define IL_V4ID 280
# define IL_TCP 281
# define IL_SPORT 282
# define IL_DPORT 283
# define IL_TCPFL 284
# define IL_TCPSEQ 285
# define IL_TCPACK 286
# define IL_TCPOFF 287
# define IL_TCPWIN 288
# define IL_TCPSUM 289
# define IL_TCPURP 290
# define IL_TCPOPT 291
# define IL_TCPO_NOP 292
# define IL_TCPO_EOL 293
# define IL_TCPO_MSS 294
# define IL_TCPO_WSCALE 295
# define IL_TCPO_TS 296
# define IL_UDP 297
# define IL_UDPLEN 298
# define IL_UDPSUM 299
# define IL_ICMP 300
# define IL_ICMPTYPE 301
# define IL_ICMPCODE 302
# define IL_SEND 303
# define IL_VIA 304
# define IL_ARP 305
# define IL_DEFROUTER 306
# define IL_SUM 307
# define IL_OFF 308
# define IL_LEN 309
# define IL_V4ADDR 310
# define IL_OPT 311
# define IL_DATA 312
# define IL_DLEN 313
# define IL_DVALUE 314
# define IL_DFILE 315
# define IL_IPO_NOP 316
# define IL_IPO_RR 317
# define IL_IPO_ZSU 318
# define IL_IPO_MTUP 319
# define IL_IPO_MTUR 320
# define IL_IPO_EOL 321
# define IL_IPO_TS 322
# define IL_IPO_TR 323
# define IL_IPO_SEC 324
# define IL_IPO_LSRR 325
# define IL_IPO_ESEC 326
# define IL_IPO_SATID 327
# define IL_IPO_SSRR 328
# define IL_IPO_ADDEXT 329
# define IL_IPO_VISA 330
# define IL_IPO_IMITD 331
# define IL_IPO_EIP 332
# define IL_IPO_FINN 333
# define IL_IPO_SECCLASS 334
# define IL_IPO_CIPSO 335
# define IL_IPO_ENCODE 336
# define IL_IPS_RESERV4 337
# define IL_IPS_TOPSECRET 338
# define IL_IPS_SECRET 339
# define IL_IPS_RESERV3 340
# define IL_IPS_CONFID 341
# define IL_IPS_UNCLASS 342
# define IL_IPS_RESERV2 343
# define IL_IPS_RESERV1 344
# define IL_ICMP_ECHOREPLY 345
# define IL_ICMP_UNREACH 346
# define IL_ICMP_UNREACH_NET 347
# define IL_ICMP_UNREACH_HOST 348
# define IL_ICMP_UNREACH_PROTOCOL 349
# define IL_ICMP_UNREACH_PORT 350
# define IL_ICMP_UNREACH_NEEDFRAG 351
# define IL_ICMP_UNREACH_SRCFAIL 352
# define IL_ICMP_UNREACH_NET_UNKNOWN 353
# define IL_ICMP_UNREACH_HOST_UNKNOWN 354
# define IL_ICMP_UNREACH_ISOLATED 355
# define IL_ICMP_UNREACH_NET_PROHIB 356
# define IL_ICMP_UNREACH_HOST_PROHIB 357
# define IL_ICMP_UNREACH_TOSNET 358
# define IL_ICMP_UNREACH_TOSHOST 359
# define IL_ICMP_UNREACH_FILTER_PROHIB 360
# define IL_ICMP_UNREACH_HOST_PRECEDENCE 361
# define IL_ICMP_UNREACH_PRECEDENCE_CUTOFF 362
# define IL_ICMP_SOURCEQUENCH 363
# define IL_ICMP_REDIRECT 364
# define IL_ICMP_REDIRECT_NET 365
# define IL_ICMP_REDIRECT_HOST 366
# define IL_ICMP_REDIRECT_TOSNET 367
# define IL_ICMP_REDIRECT_TOSHOST 368
# define IL_ICMP_ECHO 369
# define IL_ICMP_ROUTERADVERT 370
# define IL_ICMP_ROUTERSOLICIT 371
# define IL_ICMP_TIMXCEED 372
# define IL_ICMP_TIMXCEED_INTRANS 373
# define IL_ICMP_TIMXCEED_REASS 374
# define IL_ICMP_PARAMPROB 375
# define IL_ICMP_PARAMPROB_OPTABSENT 376
# define IL_ICMP_TSTAMP 377
# define IL_ICMP_TSTAMPREPLY 378
# define IL_ICMP_IREQ 379
# define IL_ICMP_IREQREPLY 380
# define IL_ICMP_MASKREQ 381
# define IL_ICMP_MASKREPLY 382
# define IL_ICMP_SEQ 383
# define IL_ICMP_ID 384
# define IL_ICMP_OTIME 385
# define IL_ICMP_RTIME 386
# define IL_ICMP_TTIME 387
