/*
 * Copyright (C) 1993-1998 by Darren Reed.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and due credit is given
 * to the original author and the contributors.
 *
 * @(#)ipl.h	1.21 6/5/96
 */

#ifndef	__IPL_H__
#define	__IPL_H__

#define	IPL_VERSION	"IP Filter v3.2.11beta4"

#endif
